
import timeit

def autotimeit(stmt, setup='pass', repeat=3, mintime=0.2):
    timer = timeit.Timer(stmt, setup)
    number, time1 = autoscaler(timer, mintime)
    time2 = timer.repeat(repeat=repeat-1, number=number)
    return min(time2 + [time1]) / number

def autoscaler(timer, mintime):
    number = 1
    for i in range(12):
        time = timer.timeit(number)
        if time > mintime:
            return number, time 
        number *= 10
    raise RuntimeError('function is too fast to test')
    



"nanequal template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["nanequal"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = 'bool'
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
                              np.ndarray[np.DTYPE_t, ndim=NDIM] b):
    "Check whether two arrays are equal, ignoring NaNs, in NDIMd array with dtype=DTYPE along axis=AXIS."
    cdef int f = 1
    cdef np.DTYPE_t ai
"""

loop = {}
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        f = 1
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            bi = b[INDEXALL]
            if ai != bi and ai == ai and bi == bi:
                y[INDEXPOP] = 0
                f = 0
                break
        if f == 1:
            y[INDEXPOP] = 1
    return y
"""
loop[3] = """\
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            f = 1
            for iINDEX2 in range(nINDEX2):
                ai = a[INDEXALL]
                bi = b[INDEXALL]
                if ai != bi and ai == ai and bi == bi:
                    y[INDEXPOP] = 0
                    f = 0
                    break
            if f == 1:
                y[INDEXPOP] = 1
    return y
"""

floats['loop'] = loop

# Float dtypes (axis=None) --------------------------------------------------

floats_None = deepcopy(floats)
floats_None['axisNone'] = True

loop = {}
loop[1] = """\
    for iINDEX0 in range(nINDEX0):
        ai = a[INDEXALL]
        bi = b[INDEXALL]
        if ai != bi and ai == ai and bi == bi:
            return np.bool_(False)
    return np.bool_(True)
"""
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            bi = b[INDEXALL]
            if ai != bi and ai == ai and bi == bi:
                return np.bool_(False)
    return np.bool_(True)
"""
loop[3] = """\
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            for iINDEX2 in range(nINDEX2):
                ai = a[INDEXALL]
                bi = b[INDEXALL]
                if ai != bi and ai == ai and bi == bi:
                    return np.bool_(False)
    return np.bool_(True)
"""

floats_None['loop'] = loop

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 

loop = {}
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        f = 1
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            bi = b[INDEXALL]
            if ai != bi:
                y[INDEXPOP] = 0
                f = 0
                break
        if f == 1:
            y[INDEXPOP] = 1
    return y
"""
loop[3] = """\
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            f = 1
            for iINDEX2 in range(nINDEX2):
                ai = a[INDEXALL]
                bi = b[INDEXALL]
                if ai != bi:
                    y[INDEXPOP] = 0
                    f = 0
                    break
            if f == 1:
                y[INDEXPOP] = 1
    return y
"""

ints['loop'] = loop


# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "nanequal"
slow['signature'] = "arr1, arr2"
slow['func'] = "bn.slow.nanequal(arr1, arr2, axis=AXIS)"

# Template ------------------------------------------------------------------

nanequal = {}
nanequal['name'] = 'nanequal'
nanequal['is_reducing_function'] = True
nanequal['cdef_output'] = True
nanequal['slow'] = slow
nanequal['templates'] = {}
nanequal['templates']['float'] = floats
nanequal['templates']['float_None'] = floats_None
nanequal['templates']['int'] = ints
nanequal['pyx_file'] = 'func/%sbit/nanequal.pyx'

nanequal['main'] = '''"nanequal auto-generated from template"

def nanequal(arr1, arr2, axis=None):
    """
    Test whether two array are equal along a given axis, ignoring NaNs.

    Returns single boolean unless `axis` is not ``None``.

    Parameters
    ----------
    arr1 : array_like
        First input array. If `arr` is not an array, a conversion is attempted.
    arr2 : array_like
        Second input array
    axis : {int, None}, optional
        Axis along which arrays are compared. The default (`axis` = ``None``)
        is to compare flattened arrays. `axis` may be
        negative, in which case it counts from the last to the first axis.

    Returns
    -------
    y : bool or ndarray
        A new boolean or `ndarray` is returned.

    See also
    --------
    bottlechest.nancmp: Compare two arrays, ignoring NaNs

    Examples -- TODO: PROVIDE EXAMPLES!
    --------
    >>> bn.nanequal(1)
    False
    >>> bn.nanequal(np.nan)
    True
    >>> bn.nanequal([1, np.nan])
    True
    >>> a = np.array([[1, 4], [1, np.nan]])
    >>> bn.nanequal(a)
    True
    >>> bn.nanequal(a, axis=0)
    array([False,  True], dtype=bool)    

    """
    func, arr1, arr2 = nanequal_selector(arr1, arr2, axis)
    return func(arr1, arr2)

def nanequal_selector(arr1, arr2, axis):
    """
    Return nanequal function and array that matches `arr` and `axis`.
    
    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in bn.nanequal()
    is in checking that `axis` is within range, converting `arr` into an
    array (if it is not already an array), and selecting the function to use.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr1 : array_like
        First input array. If `arr` is not an array, a conversion is attempted.
    arr2 : array_like
        Second input array
    axis : {int, None}, optional
        Axis along which arrays are compared. The default (`axis` = ``None``)
        is to compare flattened arrays. `axis` may be
        negative, in which case it counts from the last to the first axis.
    
    Returns
    -------
    func : function
        The nanequal function that matches the number of dimensions and
        dtype of the input array and the axis.
    a1 : ndarray
        If the input array `arr1` is not a ndarray, then `a1` will contain the
        result of converting `arr1` into a ndarray.
    a2 : ndarray
        Equivalent for arr2.

    Examples   TODO: PROVIDE EXAMPLES
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, 2.0, 3.0])
    
    Obtain the function needed to determine if there are any NaN in `arr`:

    >>> func, a = bn.func.nanequal_selector(arr, axis=0)
    >>> func
    <function nanequal_1d_float64_axisNone>
    
    Use the returned function and array to determine if there are any
    NaNs:
    
    >>> func(a)
    False

    """
    cdef np.ndarray a1, a2
    if type(arr1) is np.ndarray:
        a1 = arr1
    else:    
        a1 = np.array(arr1, copy=False)
    if type(arr2) is np.ndarray:
        a2 = arr2
    else:
        a2 = np.array(arr2, copy=False)
    cdef int ndim = PyArray_NDIM(a1)
    cdef int ndim2 = PyArray_NDIM(a1)
    if ndim != ndim2:
        raise ValueError("arrays have different dimensions, %i != %i" %
                         (ndim, ndim2))
    cdef int dtype = PyArray_TYPE(a1)
    cdef np.npy_intp *dim1, *dim2
    cdef int i
    cdef tuple key = (ndim, dtype, axis)
    if dtype == PyArray_TYPE(a2):
        dim1 = PyArray_DIMS(a1)
        dim2 = PyArray_DIMS(a2)
        for i in range(ndim):
            if dim1[i] != dim2[i]:
                raise ValueError("shape mismatch");
        if (axis is not None) and (axis < 0):
            axis += ndim
        try:
            func = nanequal_dict[key]
            return func, a1, a2
        except KeyError:
            pass

    if axis is not None:
        if (axis < 0) or (axis >= ndim):
            raise ValueError("axis(=%d) out of bounds" % axis)
    try:
        func = nanequal_slow_dict[axis]
    except KeyError:
        tup = (str(ndim), str(a1.dtype), str(axis))
        raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a1, a2
'''   



"Test list input."

# For support of python 2.5
from __future__ import with_statement

import numpy as np
from numpy.testing import (assert_equal, assert_array_equal, assert_raises,
                           assert_array_almost_equal, assert_almost_equal)
import bottlechest as bn

# ---------------------------------------------------------------------------
# Check that functions can handle list input

def lists():
    "Iterator that yields lists to use for unit testing."
    ss = {}
    ss[1] = {'size':  4, 'shapes': [(4,)]}
    ss[2] = {'size':  6, 'shapes': [(1,6), (2,3)]}
    ss[3] = {'size':  6, 'shapes': [(1,2,3)]}
    ss[4] = {'size': 24, 'shapes': [(1,2,3,4)]}  # Unaccelerated
    for ndim in ss:
        size = ss[ndim]['size']
        shapes = ss[ndim]['shapes']
        a = np.arange(size)
        for shape in shapes:
            a = a.reshape(shape)
            yield a.tolist()

def unit_maker(func, func0, args=tuple()):
    "Test that bn.xxx gives the same output as bn.slow.xxx for list input."
    msg = '\nfunc %s | input %s | shape %s\n'
    msg += '\nInput array:\n%s\n'
    for i, arr in enumerate(lists()):
        argsi = tuple([list(arr)] + list(args))
        actual = func(*argsi)
        desired = func0(*argsi)
        tup = (func.__name__, 'a'+str(i), str(np.array(arr).shape), arr)
        err_msg = msg % tup
        assert_array_almost_equal(actual, desired, err_msg=err_msg)

def test_nansum():
    "Test nansum."
    yield unit_maker, bn.nansum, bn.slow.nansum

def test_nanmax():
    "Test nanmax."
    yield unit_maker, bn.nanmax, bn.slow.nanmax

def test_nanargmin():
    "Test nanargmin."
    yield unit_maker, bn.nanargmin, bn.slow.nanargmin

def test_nanargmax():
    "Test nanargmax."
    yield unit_maker, bn.nanargmax, bn.slow.nanargmax

def test_nanmin():
    "Test nanmin."
    yield unit_maker, bn.nanmin, bn.slow.nanmin

def test_nanmean():
    "Test nanmean."
    yield unit_maker, bn.nanmean, bn.slow.nanmean

def test_nanstd():
    "Test nanstd."
    yield unit_maker, bn.nanstd, bn.slow.nanstd

def test_nanvar():
    "Test nanvar."
    yield unit_maker, bn.nanvar, bn.slow.nanvar

def test_median():
    "Test median."
    yield unit_maker, bn.median, bn.slow.median

def test_nanmedian():
    "Test nanmedian."
    yield unit_maker, bn.nanmedian, bn.slow.nanmedian
   
def test_rankdata():
    "Test rankdata."
    yield unit_maker, bn.rankdata, bn.slow.rankdata

def test_nanrankdata():
    "Test nanrankdata."
    yield unit_maker, bn.nanrankdata, bn.slow.nanrankdata

def test_ss():
    "Test ss."
    yield unit_maker, bn.ss, bn.slow.ss

def test_nn():
    "Test nn."
    a = [[1, 2], [3, 4]]
    a0 = [1, 2]
    assert_equal(bn.nn(a, a0), bn.slow.nn(a, a0))

def test_anynan():
    "Test anynan."
    yield unit_maker, bn.anynan, bn.slow.anynan

def test_allnan():
    "Test allnan."
    yield unit_maker, bn.allnan, bn.slow.allnan




"replace template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["replace"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]


# Float dtypes (axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = True
floats['force_output_dtype'] = False
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
    double old, double new):
    "replace (inplace) specified elements of NDIMd array of dtype=DTYPE."
    cdef np.DTYPE_t ai
"""     

loop = {}
loop[1] = """\
    if old == old:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai == old:
                a[INDEXALL] = new
    else:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai != ai:
                a[INDEXALL] = new
"""
loop[2] = """\
    if old == old:
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai == old:
                    a[INDEXALL] = new
    else:
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    a[INDEXALL] = new
"""
floats['loop'] = loop

# Int dtypes (axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 
ints['axisNone'] = True

ints['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
    double old, double new):
    "replace (inplace) specified elements of NDIMd array of dtype=DTYPE."
    cdef np.DTYPE_t asum = 0, ai
    cdef np.DTYPE_t oldint, newint
"""

loop = {}
loop[1] = """\
    if old == old:
        oldint = <np.DTYPE_t>old
        newint = <np.DTYPE_t>new
        if oldint != old:
            raise ValueError("Cannot safely cast `old` to int.")
        if newint != new:
            raise ValueError("Cannot safely cast `new` to int.")
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai == oldint:
                a[INDEXALL] = newint
"""
loop[2] = """\
    if old == old:
        oldint = <np.DTYPE_t>old
        newint = <np.DTYPE_t>new
        if oldint != old:
            raise ValueError("Cannot safely cast `old` to int.")
        if newint != new:
            raise ValueError("Cannot safely cast `new` to int.")
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai == oldint:
                    a[INDEXALL] = newint
"""
ints['loop'] = loop

# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "replace"
slow['signature'] = "arr, old, new"
slow['func'] = "bn.slow.replace(arr, old, new)"

# Template ------------------------------------------------------------------

replace = {}
replace['name'] = 'replace'
replace['is_reducing_function'] = False
replace['cdef_output'] = False
replace['slow'] = slow
replace['templates'] = {}
replace['templates']['float_None'] = floats
replace['templates']['int_None'] = ints
replace['pyx_file'] = 'func/%sbit/replace.pyx'

replace['main'] = '''"replace auto-generated from template"

def replace(arr, old, new):
    """
    Replace (inplace) given scalar values of an array with new values.

    The equivalent numpy function:
        
        arr[arr==old] = new
        
    Or in the case where old=np.nan:

        arr[np.isnan(old)] = new

    Parameters
    ----------
    arr : numpy.ndarray
        The input array, which is also the output array since this functions
        works inplace.
    old : scalar
        All elements in `arr` with this value will be replaced by `new`.
    new : scalar
        All elements in `arr` with a value of `old` will be replaced by `new`.

    Returns
    -------
    None, the operation is inplace. 
    
    Examples
    --------
    Replace zero with 3 (note that the input array is modified):

    >>> a = np.array([1, 2, 0])
    >>> bn.replace(a, 0, 3)
    >>> a
    array([1, 2, 3])

    Replace np.nan with 0:

    >>> a = np.array([1, 2, np.nan])
    >>> bn.replace(a, np.nan, 0)
    >>> a
    array([ 1.,  2.,  0.])
    
    """
    func = replace_selector(arr)
    return func(arr, old, new)

def replace_selector(arr):
    """
    Return replace function and array that matches `arr`.
    
    Under the hood Bottleneck uses a separate replace() Cython function for
    each combination of ndim and dtype. A lot of the overhead in bn.replace()
    is inselecting the low level function to use.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using this function.

    Parameters
    ----------
    arr : numpy.ndarray
        Input array.
    
    Returns
    -------
    func : function
        The replace() function that matches the number of dimensions and dtype
        of the input array.

    Examples
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, np.nan, 3.0])
    
    Obtain the function needed to replace values in `arr`:

    >>> func = bn.func.replace_selector(arr)
    >>> func
    <function replace_1d_float64_axisNone>

    Use the returned function to replace NaN with zero:

    >>> func(arr, np.nan, 0)
    >>> arr
    array([ 1.,  0.,  3.])

    """
    axis = None
    if type(arr) is not np.ndarray:
        # replace works in place so input must be an array, not (e.g.) a list
        raise TypeError("`arr` must be a numpy array.")
    cdef int ndim = PyArray_NDIM(arr)
    cdef int dtype = PyArray_TYPE(arr)
    cdef tuple key = (ndim, dtype, axis)
    try:
        func = replace_dict[key]
    except KeyError:
        try:
            func = replace_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(arr.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func
'''   



"Test functions."

# For support of python 2.5
from __future__ import with_statement

import numpy as np
from numpy.testing import assert_equal
nan = np.nan
import bottlechest as bn


def arrays(dtypes=bn.dtypes, nans=True):
    "Iterator that yield arrays to use for unit testing."
    ss = {}
    ss[1] = {'size':  4, 'shapes': [(4,)]} 
    ss[2] = {'size':  6, 'shapes': [(2,3)]}
    ss[3] = {'size':  6, 'shapes': [(1,2,3)]}
    for ndim in ss:
        size = ss[ndim]['size']
        shapes = ss[ndim]['shapes']
        for dtype in dtypes:
            a = np.arange(size, dtype=dtype)
            for shape in shapes:
                a = a.reshape(shape)
                yield a
            if issubclass(a.dtype.type, np.inexact): 
                if nans:
                    for i in range(a.size):
                        a.flat[i] = np.nan
                        yield a
                for i in range(a.size):
                    a.flat[i] = np.inf
                    yield a

def unit_maker(func, nans=True):
    "Test that bn.xxx gives the same output as np.xxx."
    msg =  "\nInput array modifed by %s.\n\n"
    msg += "input array before:\n%s\nafter:\n%s\n"
    for i, arr in enumerate(arrays(nans=nans)):
        for axis in list(range(-arr.ndim, arr.ndim)) + [None]:
            with np.errstate(invalid='ignore'):
                arr1 = arr.copy()
                arr2 = arr.copy()
                if ('move_' in func.__name__) or ('sort' in func.__name__):
                    if axis is None:
                        continue
                    actual = func(arr1, 1, axis=axis)
                else:
                    try:
                        actual = func(arr1, axis=axis)
                    except:
                        continue
                assert_equal(arr1, arr2, msg % (func.__name__, arr1, arr2))

def test_modification():
    "Test for illegal inplace modification of input array"
    funcs = [bn.nansum,
             bn.nanmax,
             bn.nanargmin,
             bn.nanargmax,
             bn.nanmin, 
             bn.nanmean,
             bn.nanstd,
             bn.nanvar,
             bn.median,
             bn.nanmedian,
             bn.rankdata,
             bn.nanrankdata,
             bn.ss,
             bn.anynan,
             bn.allnan,
            ]
    for func in funcs:
        yield unit_maker, func
